@@ -1,21 +1,23 @@
 #!/usr/bin/env python3
 # License: GPL v3 Copyright: 2016, Kovid Goyal <kovid at kovidgoyal.net>
 
+import os
 import shlex
 import sys
-from typing import Generator, List, Optional, Union, TYPE_CHECKING
+from contextlib import suppress
+from functools import partial
+from typing import TYPE_CHECKING, Callable, Generator, Iterator, List, Mapping, Optional, Tuple, Union
 
 from .cli_stub import CLIOptions
-from .options.utils import to_layout_names, window_size
-from .constants import kitty_exe
 from .layout.interface import all_layouts
 from .options.types import Options
+from .options.utils import resize_window, to_layout_names, window_size
 from .os_window_size import WindowSize, WindowSizeData, WindowSizes
 from .typing import SpecialWindowInstance
-from .utils import log_error, resolved_shell
-
+from .utils import cmdline_for_hold, expandvars, log_error, resolve_custom_file, resolved_shell, which
 
 if TYPE_CHECKING:
+    from .launch import LaunchSpec
     from .window import CwdRequest
 
 
@@ -27,11 +29,21 @@ def get_os_window_sizing_data(opts: Options, session: Optional['Session'] = None
     return WindowSizeData(sizes, opts.remember_window_size, opts.single_window_margin_width, opts.window_margin_width, opts.window_padding_width)
 
 
+ResizeSpec = Tuple[str, int]
+
+
+class WindowSpec:
+
+    def __init__(self, launch_spec: Union['LaunchSpec', 'SpecialWindowInstance']):
+        self.launch_spec = launch_spec
+        self.resize_spec: Optional[ResizeSpec] = None
+
+
 class Tab:
 
     def __init__(self, opts: Options, name: str):
-        from .launch import LaunchSpec
-        self.windows: List[Union[LaunchSpec, 'SpecialWindowInstance']] = []
+        self.windows: List[WindowSpec] = []
+        self.pending_resize_spec: Optional[ResizeSpec] = None
         self.name = name.strip()
         self.active_window_idx = 0
         self.enabled_layouts = opts.enabled_layouts
@@ -48,6 +60,8 @@ class Session:
         self.default_title = default_title
         self.os_window_size: Optional[WindowSizes] = None
         self.os_window_class: Optional[str] = None
+        self.os_window_state: Optional[str] = None
+        self.focus_os_window: bool = False
 
     def add_tab(self, opts: Options, name: str = '') -> None:
         if self.tabs and not self.tabs[-1].windows:
@@ -62,20 +76,43 @@ class Session:
             raise ValueError(f'{val} is not a valid layout')
         self.tabs[-1].layout = val
 
-    def add_window(self, cmd: Union[None, str, List[str]]) -> None:
+    def add_window(self, cmd: Union[None, str, List[str]], expand: Callable[[str], str] = lambda x: x) -> None:
         from .launch import parse_launch_args
+        needs_expandvars = False
         if isinstance(cmd, str):
+            needs_expandvars = True
             cmd = shlex.split(cmd)
         spec = parse_launch_args(cmd)
+        if needs_expandvars:
+            assert isinstance(cmd, list)
+            limit = len(cmd)
+            if len(spec.args):
+                with suppress(ValueError):
+                    limit = cmd.index(spec.args[0])
+            cmd = [(expand(x) if i < limit else x) for i, x in enumerate(cmd)]
+            spec = parse_launch_args(cmd)
+
         t = self.tabs[-1]
         if t.next_title and not spec.opts.window_title:
             spec.opts.window_title = t.next_title
         spec.opts.cwd = spec.opts.cwd or t.cwd
-        t.windows.append(spec)
+        t.windows.append(WindowSpec(spec))
         t.next_title = None
+        if t.pending_resize_spec is not None:
+            t.windows[-1].resize_spec = t.pending_resize_spec
+            t.pending_resize_spec = None
+
+    def resize_window(self, args: List[str]) -> None:
+        s = resize_window('resize_window', shlex.join(args))[1]
+        spec: ResizeSpec = s[0], s[1]
+        t = self.tabs[-1]
+        if t.windows:
+            t.windows[-1].resize_spec = spec
+        else:
+            t.pending_resize_spec = spec
 
     def add_special_window(self, sw: 'SpecialWindowInstance') -> None:
-        self.tabs[-1].windows.append(sw)
+        self.tabs[-1].windows.append(WindowSpec(sw))
 
     def focus(self) -> None:
         self.active_tab_idx = max(0, len(self.tabs) - 1)
@@ -90,15 +127,18 @@ class Session:
         self.tabs[-1].cwd = val
 
 
-def parse_session(raw: str, opts: Options) -> Generator[Session, None, None]:
+def parse_session(raw: str, opts: Options, environ: Optional[Mapping[str, str]] = None) -> Generator[Session, None, None]:
 
     def finalize_session(ans: Session) -> Session:
         from .tabs import SpecialWindow
         for t in ans.tabs:
             if not t.windows:
-                t.windows.append(SpecialWindow(cmd=resolved_shell(opts)))
+                t.windows.append(WindowSpec(SpecialWindow(cmd=resolved_shell(opts))))
         return ans
 
+    if environ is None:
+        environ = os.environ
+    expand = partial(expandvars, env=environ, fallback_to_os_env=False)
     ans = Session()
     ans.add_tab(opts)
     for line in raw.splitlines():
@@ -110,6 +150,8 @@ def parse_session(raw: str, opts: Options) -> Generator[Session, None, None]:
             else:
                 cmd, rest = parts
             cmd, rest = cmd.strip(), rest.strip()
+            if cmd != 'launch':
+                rest = expand(rest)
             if cmd == 'new_tab':
                 ans.add_tab(opts, rest)
             elif cmd == 'new_os_window':
@@ -119,9 +161,11 @@ def parse_session(raw: str, opts: Options) -> Generator[Session, None, None]:
             elif cmd == 'layout':
                 ans.set_layout(rest)
             elif cmd == 'launch':
-                ans.add_window(rest)
+                ans.add_window(rest, expand)
             elif cmd == 'focus':
                 ans.focus()
+            elif cmd == 'focus_os_window':
+                ans.focus_os_window = True
             elif cmd == 'enabled_layouts':
                 ans.set_enabled_layouts(rest)
             elif cmd == 'cd':
@@ -133,6 +177,10 @@ def parse_session(raw: str, opts: Options) -> Generator[Session, None, None]:
                 ans.os_window_size = WindowSizes(WindowSize(*w), WindowSize(*h))
             elif cmd == 'os_window_class':
                 ans.os_window_class = rest
+            elif cmd == 'os_window_state':
+                ans.os_window_state = rest
+            elif cmd == 'resize_window':
+                ans.resize_window(rest.split())
             else:
                 raise ValueError(f'Unknown command in session file: {cmd}')
     yield finalize_session(ans)
@@ -140,9 +188,10 @@ def parse_session(raw: str, opts: Options) -> Generator[Session, None, None]:
 
 class PreReadSession(str):
 
-    def __new__(cls, val: str) -> 'PreReadSession':
+    def __new__(cls, val: str, associated_environ: Mapping[str, str]) -> 'PreReadSession':
         ans: PreReadSession = str.__new__(cls, val)
         ans.pre_read = True  # type: ignore
+        ans.associated_environ = associated_environ  # type: ignore
         return ans
 
 
@@ -152,20 +201,25 @@ def create_sessions(
     special_window: Optional['SpecialWindowInstance'] = None,
     cwd_from: Optional['CwdRequest'] = None,
     respect_cwd: bool = False,
-    default_session: Optional[str] = None
-) -> Generator[Session, None, None]:
+    default_session: Optional[str] = None,
+) -> Iterator[Session]:
     if args and args.session:
-        if isinstance(args.session, PreReadSession):
-            session_data = '' + str(args.session)
+        if args.session == "none":
+            default_session = "none"
         else:
-            if args.session == '-':
-                f = sys.stdin
+            environ: Optional[Mapping[str, str]] = None
+            if isinstance(args.session, PreReadSession):
+                session_data = '' + str(args.session)
+                environ = args.session.associated_environ  # type: ignore
             else:
-                f = open(args.session)
-            with f:
-                session_data = f.read()
-        yield from parse_session(session_data, opts)
-        return
+                if args.session == '-':
+                    f = sys.stdin
+                else:
+                    f = open(resolve_custom_file(args.session))
+                with f:
+                    session_data = f.read()
+            yield from parse_session(session_data, opts, environ=environ)
+            return
     if default_session and default_session != 'none':
         try:
             with open(default_session) as f:
@@ -182,7 +236,8 @@ def create_sessions(
     if special_window is None:
         cmd = args.args if args and args.args else resolved_shell(opts)
         if args and args.hold:
-            cmd = [kitty_exe(), '+hold'] + cmd
+            cmd[0] = which(cmd[0]) or cmd[0]
+            cmd = cmdline_for_hold(cmd)
         from kitty.tabs import SpecialWindow
         cwd: Optional[str] = args.directory if respect_cwd and args else None
         special_window = SpecialWindow(cmd, cwd_from=cwd_from, cwd=cwd)
